#include <iostream>
#include <sstream>
#include <vector>

#include "ILUPreconditioner.h"
#include "DOFVector.h"
#include "DOFMatrix.h"
#include "DOFIterator.h"
#include "MatVecMultiplier.h"
#include "StlVector.h"

namespace AMDiS {
  
  void ILUPreconditioner::init()
  {
    FUNCNAME("ILUPreconditioner::init()");

    clock_t first = clock();

    ilu_.resize(matrix.getSize());

    for (int i = 0; i < matrix.getSize(); i++) {
      if ((*matrix[i])) {
	ilu_[i] = NEW DOFMatrix((*(*matrix[i])));
      }
    }

 
    DOFMatrix *ilu = ilu_[row];   

    int matrixSize = ilu->getRowFESpace()->getAdmin()->getUsedSize();

    // Sort all rows corresponding to their position in the matrix.
    diag_.resize(matrixSize);
    for (int i = 0; i < matrixSize; i++) {
      diag_[i] = (*ilu)[i][0].entry;
      sort((*ilu)[i].begin(), (*ilu)[i].end(), CmpMatEntryCol());
    }

    // For i = 1, ..., n - 1, Do
    for (int i = 1; i < matrixSize; i++) {

      // For k = 0, ..., i - 1 and for (i,k) \in NZ(ilu), Do
      for (int col1 = 0; col1 < static_cast<int>((*ilu)[i].size()); col1++) {
	// k is the logical index of the col1-th column in the i-th row.
	int k = (*ilu)[i][col1].col;
	if ((k < i) &&                         // due to the condition that k goes up to i - 1
	    (k >= 0) &&                        // the entry is valid
	    ((*ilu)[i][col1].entry != 0.0)) {     // the entry is non-zero, hence in NZ(ilu)
	  // Compute a_ik = a_ik / a_kk
	  (*ilu)[i][col1].entry = (*ilu)[i][col1].entry / diag_[k];

	  // For j = k + 1, ..., n - 1 and for (i,j) \in NZ(ilu), Do
	  for (int col2 = col1 + 1; col2 < static_cast<int>((*ilu)[i].size()); col2++) {
	    // j is the logical index of the col2-th column in the i-th row.
	    int j = (*ilu)[i][col2].col;

	    // Because the rows are sorted, we start at position col1 + 1, and all non-valid 
	    // entries have col < 0, we do not need to check for non-valid entries.
	    if ((*ilu)[i][col2].entry != 0.0) {
	      int physKJ = (*ilu).logToPhysIndex(k, j);
	      // If physKJ is negativ, the corresponding entry is zero and we do not need 
	      // to compute the new element.
	      if (physKJ >= 0) {
		// Compute a_ij = a_ij - a_ik * a_kj
		(*ilu)[i][col2].entry = (*ilu)[i][col2].entry - ((*ilu)[i][col1].entry * (*ilu)[k][physKJ].entry);
		// Do not forgett to update the diagonal entries, if we have computed on
		// a diagonal element.
		if (i == j) {
		  diag_[i] = (*ilu)[i][col2].entry;
		}
	      }
	    }
	  }
	}
      }      
    }

    MSG("preconditioning of the coefficient matrix needed %.5f seconds\n",
	TIME_USED(first,clock()));
  }

  void ILUPreconditioner::exit()
  {
    for (int i = 0; i < ilu_.getSize(); i++) {
      if ((*matrix[i])) {
	DELETE ilu_[i];
      }
    }
  }

  void ILUPreconditioner::precon(DOFVector<double>* x)
  { 
    DOFMatrix *ilu = ilu_[row];

    // The solution of P * x = b must be calculated in the
    // following. Note that the vector b is stored in x to
    // saving space.
    // We solve the equation with forward and backward substitution
    // of the factorization P = L * U.
 
    DOFVector<double>::Iterator itx(x, USED_DOFS);

    // Forward substitution: L * y = b
    // We use x to save the new results of y. Note that all diagonal 
    // elements of L are 1, hence we do not need to devide the term
    // by the diagonal elements.
    int i = 0;
    for (itx.reset(); !itx.end(); ++itx, ++i) {
      // y_i = b_i
      // Must not be assigned directly, since x = b and y will take
      // the results of y.
      
      ::std::vector<MatEntry> *m = &(*ilu)[i];
      
      // y_i = b_i - \sum_{1}^{i-1} ( l_{i,k} * y_k )
      for (int col = 0; col < static_cast<int>((*m).size()); col++) {
	int k = (*m)[col].col;
	if (k >= 0) {
	  if (k < i) {
	    (*itx) -= (*m)[col].entry * (*x)[k];
	  } else {
	    // If for an entry k < i does not hold, it will be also false
	    // for all other elements in th column, because the columns 
	    // are sorted.
	    break;
	  }
	}
      }
    }
    
    
    // Backward substitution: R * x = y
    do {
      itx--;
      i--;

      ::std::vector<MatEntry> *m = &(*ilu)[i];
      
      // x_i = y_i - \sum_{i + 1}^(n} ( r_{i,k} * x_k )
      // The sum is build backwards, since the directon of calculation does not 
      // matter for the result, but we can break the search if k > i does not 
      // hold (the cols of ilu_ are sorted!) and so make the calculation faster.
      for (int col = static_cast<int>((*m).size()) - 1; col >= 0; col--) {
 	int k = (*m)[col].col;
  	if (k > i) {
  	  (*itx) -= (*m)[col].entry * (*x)[k];
  	} else {
	  break;
	}
      }

      // x_i = x_i / r_{i,i}
      (*itx) /= diag_[i];
      
    } while (!itx.begin());
  }









  /////////////////////////////

  void ILUPreconditionerStd::init()
  {
    FUNCNAME("ILUPreconditionerStd::init()");

    clock_t first = clock();

    ilu_ = *matrix;

    int matrixSize = ilu_.size();

//     ilu_.resize(matrixSize);

//     for (int i = 0; i < matrixSize; i++) {
//       ilu_[i] = (*matrix
//     }

    ::std::vector< ::std::vector<MatEntry> > *ilu = &ilu_;



    // Sort all rows corresponding to their position in the matrix.
    diag_.resize(matrixSize);
    for (int i = 0; i < matrixSize; i++) {
      diag_[i] = (*ilu)[i][0].entry;
      sort((*ilu)[i].begin(), (*ilu)[i].end(), CmpMatEntryCol());
    }

    // For i = 1, ..., n - 1, Do
    for (int i = 1; i < matrixSize; i++) {

      // For k = 0, ..., i - 1 and for (i,k) \in NZ(ilu), Do
      for (int col1 = 0; col1 < static_cast<int>((*ilu)[i].size()); col1++) {
	// k is the logical index of the col1-th column in the i-th row.
	int k = (*ilu)[i][col1].col;
	if ((k < i) &&                         // due to the condition that k goes up to i - 1
	    (k >= 0) &&                        // the entry is valid
	    ((*ilu)[i][col1].entry != 0.0)) {     // the entry is non-zero, hence in NZ(ilu)
	  // Compute a_ik = a_ik / a_kk
	  (*ilu)[i][col1].entry = (*ilu)[i][col1].entry / diag_[k];

	  // For j = k + 1, ..., n - 1 and for (i,j) \in NZ(ilu), Do
	  for (int col2 = col1 + 1; col2 < static_cast<int>((*ilu)[i].size()); col2++) {
	    // j is the logical index of the col2-th column in the i-th row.
	    int j = (*ilu)[i][col2].col;

	    // Because the rows are sorted, we start at position col1 + 1, and all non-valid 
	    // entries have col < 0, we do not need to check for non-valid entries.
	    if ((*ilu)[i][col2].entry != 0.0) {
	      //	      int physKJ = (*ilu).logToPhysIndex(k, j);
	      int physKJ = logToPhysIndex(ilu, k, j);

	      // If physKJ is negativ, the corresponding entry is zero and we do not need 
	      // to compute the new element.
	      if (physKJ >= 0) {
		// Compute a_ij = a_ij - a_ik * a_kj
		(*ilu)[i][col2].entry = (*ilu)[i][col2].entry - ((*ilu)[i][col1].entry * (*ilu)[k][physKJ].entry);
		// Do not forgett to update the diagonal entries, if we have computed on
		// a diagonal element.
		if (i == j) {
		  diag_[i] = (*ilu)[i][col2].entry;
		}
	      }
	    }
	  }
	}
      }      
    }

    MSG("preconditioning of the coefficient matrix needed %.5f seconds\n",
	TIME_USED(first,clock()));
  }

  void ILUPreconditionerStd::exit()
  {
  }

    void ILUPreconditionerStd::precon(::std::vector<double> *x)
    { 
    ::std::vector< ::std::vector<MatEntry> > *ilu = &ilu_;

    // The solution of P * x = b must be calculated in the
    // following. Note that the vector b is stored in x to
    // saving space.
    // We solve the equation with forward and backward substitution
    // of the factorization P = L * U.
 
      //    DOFVector<double>::Iterator itx(x, USED_DOFS);

    ::std::vector<double>::iterator itx;

    // Forward substitution: L * y = b
    // We use x to save the new results of y. Note that all diagonal 
    // elements of L are 1, hence we do not need to devide the term
    // by the diagonal elements.
    int i = 0;
    for (itx = x->begin(); itx != x->end(); ++itx, ++i) {
      // y_i = b_i
      // Must not be assigned directly, since x = b and y will take
      // the results of y.
      
      ::std::vector<MatEntry> *m = &(*ilu)[i];
      
      // y_i = b_i - \sum_{1}^{i-1} ( l_{i,k} * y_k )
      for (int col = 0; col < static_cast<int>((*m).size()); col++) {
	int k = (*m)[col].col;
	if (k >= 0) {
	  if (k < i) {
	    (*itx) -= (*m)[col].entry * (*x)[k];
	  } else {
	    // If for an entry k < i does not hold, it will be also false
	    // for all other elements in th column, because the columns 
	    // are sorted.
	    break;
	  }
	}
      }
    }
    
    
    // Backward substitution: R * x = y
    do {
      itx--;
      i--;

      ::std::vector<MatEntry> *m = &(*ilu)[i];
      
      // x_i = y_i - \sum_{i + 1}^(n} ( r_{i,k} * x_k )
      // The sum is build backwards, since the directon of calculation does not 
      // matter for the result, but we can break the search if k > i does not 
      // hold (the cols of ilu_ are sorted!) and so make the calculation faster.
      for (int col = static_cast<int>((*m).size()) - 1; col >= 0; col--) {
 	int k = (*m)[col].col;
  	if (k > i) {
  	  (*itx) -= (*m)[col].entry * (*x)[k];
  	} else {
	  break;
	}
      }

      // x_i = x_i / r_{i,i}
      (*itx) /= diag_[i];
      
    } while (itx != x->begin());
  }

}
